from transformers import AutoModelForSequenceClassification
import torch
import torch.nn as nn
from torch.nn import CrossEntropyLoss, MSELoss
from transformers.modeling_outputs import SequenceClassifierOutput
from torch import linalg as LA
import torch.nn.functional as F
from transformers.file_utils import ModelOutput
from dataclasses import dataclass
from typing import Optional, Tuple
import gc
import inspect
import warnings
@dataclass
class CoTrainModelOutput(ModelOutput):
    
    loss: Optional[torch.FloatTensor]  = None
    logits_m1: torch.FloatTensor = None
    logits_m2: torch.FloatTensor = None

@dataclass
class TriTrainModelOutput(ModelOutput):
    
    loss: Optional[torch.FloatTensor]  = None
    logits_m1: torch.FloatTensor = None
    logits_m2: torch.FloatTensor = None
    logits_m3: torch.FloatTensor = None

def add_signature_from(base):
    
    def decorator(derived):
        
        #These keys are irrelevant for hyperparameters but create duplicate key issues. 
        ignore_keys=['ssl_model_type', 'num_models']
        
        base_params = [v for k, v in inspect.signature(base).parameters.items() if k not in ignore_keys]
        der_params = list(inspect.signature(derived).parameters.values())
        
        ##removes the *args and **kwargs from the signature
        new_params = set(der_params[:-2]+ base_params)
        derived.__signature__ = inspect.signature(derived).replace(parameters=new_params)
        return derived
    
    return decorator 

class BaseModelClass(nn.Module):
    
    """
    Base class for all model with single pretrained model, but might have multiple classifier layers.

    Args:

    model_name (:obj:`str` or :obj:`os.PathLike`): "pretrained_model_name_or_path" 
    in ~transformers.PreTrainedModel, please refer to its documentation for further information. 
    
    (i) In this case of a string, the `model id` of a pretrained model hosted inside a model 
    repo on huggingface.co.
    
    (ii) It could also be address of saved pretrained model. 
    
    supervised_run (:obj:`bool`): If the model is taken from the supervised run or not. 
    In that case transformer_model_name is the path to the saved model. 
    
    num_labels (:obj:`int`): number of labels to be classified.
    
    classifier_dropout (:obj:`float`): dropout probability of the classifier layers.
        
    num_models (:obj:`int`): number of models, i.e. number of classifier layers 
    (only set by the sub classes).
    
    ssl_model_type (:obj:`str`): semi supervised learning model type (only set by the sub classes).    
    
    """

    def __init__(self, model_name='albert-base-v2', supervised_run=False, num_labels=2, classifier_dropout=0.1,
                 num_models=1, ssl_model_type=None):
        super().__init__()

        self.type_ = ssl_model_type
        self.num_labels = num_labels
        self.num_models = num_models
        
        if supervised_run:
            self.pretrained_model = AutoModelForSequenceClassification.from_pretrained(model_name)
            
        else:
            self.pretrained_model = AutoModelForSequenceClassification.from_pretrained(model_name, num_labels=num_labels)                
            
        if num_models>1:
            # if number of classifiers are more than one then the AutoModels classifiers is removed and
            # reinitialized here.  
            del self.pretrained_model.classifier
            gc.collect()
            
            self.config = self.pretrained_model.config
            self.hidden_size = self.pretrained_model.config.hidden_size
            self.num_labels = self.config.num_labels
            self.classifiers = nn.ModuleList([nn.Linear(self.hidden_size, self.num_labels) for _ in range(num_models)])
            self.dropouts = nn.ModuleList([nn.Dropout(classifier_dropout) for _ in range(num_models)])
            self.softmaxs = nn.ModuleList([nn.Softmax(1) for _ in range(num_models)])
            self.__init__weights()
                
    def __init__weights(self):
        
        """
        Reinitialization of __init__weights function of 'pretrained_model' for 
        extra added classifier layers.        
        
        """
        for ind in range(self.num_models):
            self.pretrained_model._init_weights(self.classifiers[ind])
    
    def simple_forward_with_prob_logits(self, classifier_num=0, **kwargs):
        """
        This function first changes the pointer of the pretrained_model to the 
        one of the classifier defined in this class. Then applies softmax to it 
        and thus converts it to probability logits.
        
        Args:
        
        classifier_num: Index of the classifier to be used.

        kwargs: Arguments from pretrained_model.forward. 
        
        Return:
        
        logits ( torch.FloatTensor): probability logits.        
        
        """
        self.pretrained_model.classifier = self.classifiers[classifier_num]
        
        outputs = self.pretrained_model(**kwargs)

        logits = self.softmaxs[classifier_num](outputs.logits)
        
        return logits

@add_signature_from(BaseModelClass)    
class PiModel(BaseModelClass):
    
    """
    Implementation of pi model from <https://arxiv.org/abs/1610.02242>.

    Args:
        
    unsup_weight (:obj:`float`): Initial value of the weight of the unsupervised 
    loss component. Its value is controlled by unsupervised weight scheduler.  
    
    kwargs: remaining dictionary of keyword arguments from the BaseModelClass.
    
    """

    def __init__(self, unsup_weight=0, *args, **kwargs):
        super().__init__(*args, **kwargs, ssl_model_type="PiModel")
        
        self.unsup_weight=unsup_weight
        self.simple_forward = self.pretrained_model.forward
        
    def forward(self, **kwargs):
        
        """ 
        Implementation of forward function calculating the 
        semi supervised loss. Mixing of the labeled and unlabeled examples 
        in a single batch is not allowed. 
        
        Args:
        
        kwargs: Arguments from pretrained_model.forward.
        
        Return:
        transformers.modeling_outputs.SequenceClassifierOutput object with the information 
        of logits and loss function.
        """

        labels=kwargs.pop('labels')
    
        z1_logits = self.simple_forward( **kwargs ).logits
        z2_logits = self.simple_forward( **kwargs ).logits
        
        sup_loss = CrossEntropyLoss()
        unsup_loss = MSELoss()
        
        tot_loss=0
        if all(labels>=0):
            tot_loss = sup_loss(z1_logits.view(-1, self.num_labels), labels.view(-1))
    
        tot_loss += self.unsup_weight*unsup_loss(z1_logits, z2_logits)
        
        return SequenceClassifierOutput(
            loss=tot_loss,
            logits=z1_logits
        )
    
@add_signature_from(BaseModelClass)    
class TemporalEnsembleModel(BaseModelClass):
    
    """
    Implementation of Temporal ensemble model as introduced in <https://arxiv.org/abs/1610.02242>

    Args:
        
    alpha (:obj:`float`): memory of the last epochs. For more info please refer 
    to <https://arxiv.org/abs/1610.02242>.

    unsup_weight (:obj:`float`): initial value of weight of the unsupervised loss
    component. After setting the initial value, its value is controlled by 
    unsupervised weight scheduler. 
    
    kwargs: remaining dictionary of keyword arguments from the BaseModelClass.
    
    Class attributes:
    
     -**mini_batch_num**: keeps track of the mini_batch_num using forward method.
     
     -**logits_batchwise**: stores the logits of each batch passed through forward method.
     
     -**firstpass**: bool variable to track if its the first pass through the forward method.
     
    """

    def __init__(self, alpha=0.5, unsup_weight=0, *args, **kwargs):
        
        super().__init__( *args, **kwargs, ssl_model_type="TemporalEnsembleModel")
        
        self.unsup_weight=unsup_weight
        self.memory_logits=[]
        self.alpha=alpha
        self.mini_batch_num=0
        self.logits_batchwise=[]
        self.firstpass=True
        self.simple_forward = self.pretrained_model.forward
    
    def forward(self, **kwargs):
        
        """ 
        Implementation of forward function calculating the semi supervised loss.
        Mixing of the labeled and unlabeled examples in a single batch is not allowed. 
        
        Args:
        
        kwargs: Arguments from pretrained_model.forward.
        
        Return:
        transformers.modeling_outputs.SequenceClassifierOutput object 
        with the information of logits and the loss function.
        
        """
        
        labels = kwargs.pop("labels")
        logits = self.simple_forward(**kwargs).logits
        
        sup_loss = CrossEntropyLoss()
        unsup_loss = MSELoss()
        
        tot_loss=0
        if all(labels>=0):
            tot_loss = sup_loss(logits.view(-1, self.num_labels), labels.view(-1))
        
        if self.firstpass==False and self.training:
            tot_loss += self.unsup_weight*unsup_loss(self.memory_logits[self.mini_batch_num], logits)
            
        elif self.training:
            ##required to free the graph
            tot_loss += 0*logits.sum()
            
        if self.training:
            self.logits_batchwise.append(logits.detach().clone())
            self.mini_batch_num += 1    
        
        
        return SequenceClassifierOutput(
            loss=tot_loss,
            logits=logits
        ) 
    
    def update_memory_logits(self, t):
        
        """
        Method for updating the memory logits with the exponential average.
        
        Args:
        t (:obj:`int`): epoch value of bias normalization.
        
        """

        if self.firstpass==True:
            
            device = self.logits_batchwise[0].device
            
            self.memory_logits=[torch.zeros(self.logits_batchwise[0].shape,device=device) 
                                for _ in range(self.mini_batch_num)]
            self.firstpass=False
            
        for ind in range(self.mini_batch_num):
            
            self.memory_logits[ind] = (self.alpha*self.memory_logits[ind] + 
                                       (1-self.alpha)*self.logits_batchwise[ind])
            
            self.memory_logits[ind] /= (1-self.alpha**(t))
        
        self.mini_batch_num=0
        self.logits_batchwise[:]=[]

@add_signature_from(BaseModelClass)                
class CoTrain(BaseModelClass):
    
    """
    Implementation of Co Training as introduced in <https://www.cs.cmu.edu/~avrim/Papers/cotrain.pdf>

    Args:
        
    o_weight (:obj:`float`): Orthogonality weight for the two classifiers (or two models).
    
    kwargs: remaining dictionary of keyword arguments from the BaseModelClass.
         
    """

    def __init__(self, o_weight=0.01, ssl_model_type="CoTrain", num_models=2, *args, **kwargs):
        
        if type(self).__name__=='CoTrain':
            if ssl_model_type != 'CoTrain' or num_models!=2:
                raise RuntimeError('ssl_model_type or num_models can only be changed through a subclass.')
                
        
        super().__init__(*args, **kwargs, num_models = num_models, ssl_model_type = ssl_model_type)
        
        self.simple_forward = super().simple_forward_with_prob_logits
        self.o_weight=o_weight
        
    def forward(self, **kwargs):
        
        """ 
        Forward function only used during the evaluation of models.
        See ~trainer_utils.TrainerForCoTraining and ~transformers.Trainer for more details.  
        
        Args:
        
        kwargs: Arguments from pretrained_model.forward.
        
        Return:
        CoTrainModelOutput object with the information of logits of both models and the 
        loss function.
        
        """
        labels = kwargs.pop("labels")
        logits1 = self.simple_forward(0, **kwargs)
        logits2 = self.simple_forward(1, **kwargs)
        
        sup_loss = CrossEntropyLoss()
        
        loss1 = sup_loss(logits1.view(-1, self.num_labels), labels.view(-1))
        loss2 = sup_loss(logits2.view(-1, self.num_labels), labels.view(-1))
        
        tot_loss= (loss1 + loss2 + 
                   self.o_weight * LA.norm(torch.matmul(self.classifiers[0].weight.t(), self.classifiers[1].weight)))
        
        return CoTrainModelOutput(
            loss=tot_loss,
            logits_m1=logits1,
            logits_m2=logits2
        )


    def cotrain_forward(self, model1_batch, model2_batch):
        
        """ 
        Forward function used during training of models. 
        See ~trainer_utils.TrainerForCoTraining for more details.  
        
        Args:
        model1_batch (:obj: torch.FloatTensor) batch for model 1.
        model2_batch (:obj: torch.FloatTensor) batch for model 2.
        
        Return:
        CoTrainModelOutput object with the information of logits of both models and the 
        loss function.
        
        """


        labels1 = model1_batch.pop('labels')
        labels2 = model2_batch.pop('labels')

        logits1 = self.simple_forward(classifier_num=0, **model1_batch)
        logits2 = self.simple_forward(classifier_num=1, **model2_batch)
        
        sup_loss = CrossEntropyLoss()
        
        loss1 = sup_loss(logits1.view(-1, self.num_labels), labels1.view(-1))
        loss2 = sup_loss(logits2.view(-1, self.num_labels), labels2.view(-1))
        
        tot_loss =(loss1 + loss2 + 
                   self.o_weight * LA.norm(torch.matmul(self.classifiers[0].weight.t(), self.classifiers[1].weight)))

        return CoTrainModelOutput(
            loss=tot_loss,
            logits_m1=logits1,
            logits_m2=logits2
        )

@add_signature_from(CoTrain)  
class TriTrain(CoTrain):
    
    """
    Implementation of Tri Training(multi task TriTrain) as introduced 
    in <https://arxiv.org/abs/1804.09530>. Note: Here the implementation 
    is at only the fine tuning. The base network is to be pretrained transformer model.

    Args:
        
    kwargs: keyword arguments are the same as is for CoTrain class, except model 
    type string and number of models(num_models) as obvious with the name.
         
    """

    def __init__(self, *args, **kwargs):
        
        super().__init__( *args, ssl_model_type="TriTrain", num_models=3, **kwargs)
        
    
    def forward(self, **kwargs):
        
        """ 
        Forward function used during evaluation of trained models.
        See ~trainer_utils.TrainerForTriTraining and ~transformers.Trainer 
        for more details.  
        
        Args:
        
        kwargs: Arguments from pretrained_model.forward.
        
        Return:
        TriTrainModelOutput object with the information of logits of both 
        models and the loss function.
        
        """

        labels = kwargs["labels"]
        ct_output = super().forward(**kwargs)
        
        logits = self.simple_forward(2, **kwargs)
        
        sup_loss = CrossEntropyLoss()

        loss =  ct_output.loss + sup_loss(logits.view(-1, self.num_labels), labels.view(-1))
        
        return TriTrainModelOutput(
            loss = loss,
            logits_m1 = ct_output.logits_m1,
            logits_m2 = ct_output.logits_m2,
            logits_m3 = logits,
        )
    
    def m3_forward(self, **kwargs):

        """ 
        Forward function for model 3.
        See ~trainer_utils.TrainerForTriTraining and ~transformers.Trainer for
        more details.  
        
        Args:
        
        kwargs: Arguments from pretrained_model.forward.
        
        Return:
        TriTrainModelOutput object with the information of logits of both 
        models and the loss function.
        
        """
        labels = kwargs.pop("labels")

        logits = self.simple_forward(2, **kwargs)

        sup_loss = CrossEntropyLoss()
                
        loss = sup_loss(logits.view(-1, self.num_labels), labels.view(-1))
        
        return SequenceClassifierOutput(
            loss = loss,
            logits = logits)

    
class BaseMultiPretrained(nn.Module):
    
    """
    Base class for all models with multiple pretrained model. 

    Args:
    
    ssl_model_type (:obj:`str`): semi supervised learning model type.

    teacher_student_name (:obj:`Tuple[`str`, `str`]): A Tuple for teacher 
    and student name, respectively. "pretrained_model_name_or_path" in 
    ~transformers.PreTrainedModel, please refer to its documentation for 
    further information. 
    
    (i) In this case of a string, the `model id` of a pretrained model 
    hosted inside a model repo on huggingface.co.
    
    (ii) It could also be address of saved pretrained model. 
        
    num_labels (:obj:`int`): number of labels to be classified.
    
    student_dropout (:obj:`float`): dropout probability of the student classifier layers.

    teacher_dropout (:obj:`float`): dropout probability of the teacher classifier layers.    
        
    """
    
    def __init__(self, teacher_student_name=("albert-base-v2","albert-base-v2"), num_labels=2, teacher_dropout=None, student_dropout=None, ssl_model_type=None):
        
        super().__init__()
        
        self.teacher_model_name = teacher_student_name[0]        
        self.student_model_name = teacher_student_name[1]

        self.type_ = ssl_model_type
        self.num_labels = num_labels
        
        self.teacher = AutoModelForSequenceClassification.from_pretrained(self.teacher_model_name, num_labels = num_labels)
        
        if teacher_dropout is not None: self.teacher.dropout.p = teacher_dropout
        
        self.student = AutoModelForSequenceClassification.from_pretrained(self.student_model_name, num_labels = num_labels)
        
        if student_dropout is not None: self.student.dropout.p = student_dropout
    
    
            
@add_signature_from(BaseMultiPretrained)                
class MeanTeacher(BaseMultiPretrained):
    
    """
    Implementation of Mean Teacher as introduced in <https://arxiv.org/abs/1703.01780>

    Args:
        
    alpha (:obj:`float`): memory of the last epochs.
    
    unsup_weight: Initial unsupervised weight.
    
    Class attributes:
         
     -**firstpass**: bool variable to track if its the first pass through the forward method.
     
    """

    def __init__(self, alpha=0.5, unsup_weight=0, *args, **kwargs):
        
        super().__init__(*args, ssl_model_type='MeanTeacher', **kwargs)
        
        if self.teacher_model_name != self.student_model_name:
            warnings.warn("When using different pretrained models for mean teacher, confirm that both have similar parameters, for eg. number and size of hidden layers, attention etc.")
        
        self.apply(self.zero_teacher_weights)
        
        self.unsup_weight=unsup_weight
        self.alpha=alpha
        self.firstpass=True

    def zero_teacher_weights(self, module):
        
        """ 
        Function for zeroing the teachers weights and biases.
        
        """
        
        if hasattr(module, 'weight'):
            module.weight.data.zero_()
        
        if hasattr(module, 'bias'):
            module.bias.data.zero_()
    
        
    def forward(self, **kwargs):
        
        """ 
        Implementation of forward function calculating the semi supervised loss.
        Mixing of the labeled and unlabeled examples in a single batch is not allowed. 
        
        Args:
        
        kwargs: Arguments from pretrained_model.forward.
        
        Return:
        transformers.modeling_outputs.SequenceClassifierOutput object with the 
        information of logits and the loss function.
        
        """

        labels = kwargs.pop("labels")
        stud_logits = self.student(**kwargs).logits
        
        sup_loss = CrossEntropyLoss()
        unsup_loss = MSELoss()
        tot_loss=0
        
        if all(labels>=0):
            tot_loss = sup_loss(stud_logits.view(-1, self.num_labels), labels.view(-1))
        
        teach_logits=None
        
        if self.firstpass==False and self.training:
            
            with torch.no_grad():
                teach_logits = self.teacher(**kwargs).logits
                
            tot_loss += self.unsup_weight*unsup_loss(teach_logits, stud_logits)
        
        elif self.training:
            ##required to free the graph in the firstpass
            tot_loss += 0*stud_logits.sum()

        
        logits = teach_logits if teach_logits is not None else stud_logits
        
        return SequenceClassifierOutput(
            loss=tot_loss,
            logits=logits
        )
    

    def update_teacher_variables(self):

        """
        Function for updating teacher weights and bias. Directly
        used from <https://github.com/CuriousAI/mean-teacher>
                        
        """
        self.firstpass=False

        for teach_param, stud_param in zip(self.teacher.parameters(), self.student.parameters()):
            teach_param.data.mul_(self.alpha).add_(stud_param.data, alpha = 1 - self.alpha)        

        
@add_signature_from(BaseMultiPretrained)                
class NoisyStudent(BaseMultiPretrained):
    
    """
    Implementation of Noisy Student as introduced in <https://arxiv.org/abs/1911.04252>

    Args:
    
    kwargs: keyword arguments are the same as is for BaseMultiPretrained class,
    except model type string. Class forward initialized with teacher as the teacher is trained first. 
    
    """

    def __init__(self, *args, **kwargs):
        
        
        super().__init__(*args, ssl_model_type='NoisyStudent', **kwargs)

        self.teacher_dropout = self.teacher.dropout.p
        self.student_dropout = self.student.dropout.p
        
        if self.teacher_dropout > self.student_dropout:
            warnings.warn("For NoisyStudent to work properly, the teacher classifier dropout should be lower than student classifier dropout.")
        
        self.forward = self.student.forward



from transformers import TrainingArguments
import numpy as np
from .dataset_utils import modify_datasets, SimpleDataset, extract_keys
from transformers import Trainer
from transformers import AutoTokenizer
import warnings



# default keyword arguments for transformers.TrainingArguments for supervised and
# semisupervised models. Set_default_args changes this dictionary depending what 
# setup of args or args_sup in the train_with_ssl keyword arguments.

kwargs_args_sup = {'output_dir':"glue",
                  'evaluation_strategy':"epoch",
                  'learning_rate':2e-5,
                  'per_device_train_batch_size':16,
                  'per_device_eval_batch_size':16,
                  'num_train_epochs':10,
                  'weight_decay':0.01,
                  'metric_for_best_model':"matthews_correlation",
                  'load_best_model_at_end':True,
                  'disable_tqdm':True,
                  'no_cuda' :True}

kwargs_args = kwargs_args_sup.copy()
kwargs_args['save_steps']=np.inf

def get_default_cm():
    
    """
    Default compute metric function.
    """
    from datasets import load_metric
    metric = load_metric('glue', "cola")

    def compute_metrics(eval_pred):
        predictions, labels = eval_pred
        predictions = np.argmax(predictions, axis=1)
        return metric.compute(predictions=predictions, references=labels)
    
    return compute_metrics

def encode(dataset, model_name = 'albert-base-v2', text_column_name='sentence'):

    """
    Function for encoding the dataset using tokenizer 
    
    Args:
    
    dataset (:obj:`~datasets.DatasetDict` ): Dataset dictionary containing labeled
    and unlabeled data.
    
    model_name (:obj:`str` or :obj:`os.PathLike`): "pretrained_model_name_or_path" 
    in ~transformers.PreTrainedModel, please refer to its documentation for further 
    information. 
    
    (i) In this case of a string, the `model id` of a pretrained model 
    hosted inside a model repo on huggingface.co.
    
    (ii) It could also be address of saved pretrained model. 
    
    text_column_name (:obj:`str`): column name for where the text is.
    
    Return:
    
    encoded_dataset (:obj:`~datasets.DatasetDict` ): containing columns now 
    which are required by the forward function.
    
    tokenizer (:obj:`~transformers.PreTokenizer` ):
    
    """
    
    tokenizer = AutoTokenizer.from_pretrained(model_name, use_fast=True)
    
    def preprocess_function(examples):
        return tokenizer(examples[text_column_name], truncation=True)
    
    encoded_dataset = dataset.map(preprocess_function, batched=True)
    
    return encoded_dataset, tokenizer


def set_default_args(dataset, model_name, kwargs):
    
    """
    Function for setting the default arguments if these keywords are not provided in
    kwargs of train_with_ssl. Updates the kwargs to be used transformers.Trainer.
    
    Args:
    
    dataset (:obj:`~datasets.DatasetDict` ): Dataset dictionary containing labeled and
    unlabelled data.
    
    model_name (:obj:`str` or :obj:`os.PathLike`): "pretrained_model_name_or_path" in 
    ~transformers.PreTrainedModel, please refer to its documentation for further information. 
    
    (i) In this case of a string, the `model id` of a pretrained model hosted inside a 
    model repo on huggingface.co.
    
    (ii) It could also be address of saved pretrained model. 
    
    kwargs (:obj:`dict`): keyword arguments to be used by transformers.Trainer.    
    
    """
  
    kwargs_trn = extract_keys(Trainer, kwargs)
    
    if 'compute_metrics' not in kwargs_trn.keys():
        kwargs_trn['compute_metrics'] = get_default_cm()
        
    if 'tokenizer' not in kwargs_trn.keys():
        text_column_name = kwargs.pop('text_column_name', 'sentence')
        encoded_dataset, tokenizer = encode(dataset, model_name, text_column_name)
        kwargs_trn['tokenizer'] = tokenizer
    else:
        warnings.warn("tokenizer found. If using tokenizer, please use the encoded dataset. Like done using ~set_default_args.encode()")
        
    if 'args_ta_sup' in kwargs.keys():
        
        if isinstance( kwargs['args_ta_sup'], dict):
            kwargs_args_sup.update(kwargs['args_ta_sup'])
            del kwargs['args_ta_sup']

    if 'args_ta' in kwargs.keys():
        
        if isinstance( kwargs['args_ta'], dict):
            kwargs_args.update(kwargs['args_ta'])
            del kwargs['args_ta']
    
    kwargs.update(kwargs_trn)
    return encoded_dataset, kwargs_args
    
def get_default_ta_sup(logging_dir=''):
    """ 
    Return the TrainingArguments with the logging_dir setup for supervised model.
    """

    kwargs_args_sup['logging_dir']=logging_dir
    return TrainingArguments(**kwargs_args_sup)
    
def get_default_ta(logging_dir=''):
    """ 
    Return the TrainingArguments with the logging_dir setup for semisupervised model.
    """
    kwargs_args['logging_dir']=logging_dir
    return TrainingArguments(**kwargs_args)




import torch
from transformers import Trainer
from transformers import AutoModelForSequenceClassification,  Trainer
from .dataset_utils import modify_datasets, SimpleDataset, extract_keys
import os, shutil
import numpy as np
import gc
from collections import defaultdict, OrderedDict
import itertools
import inspect
from .default_args import get_default_ta_sup, get_default_ta, set_default_args
from functools import wraps

from .trainer_util import (
    TrainerWithUWScheduler, 
    TrainerForCoTraining, 
    TrainerForTriTraining,
    TrainerForNoisyStudent
)
from .models import (
    PiModel,
    TemporalEnsembleModel,
    CoTrain,
    TriTrain,
    MeanTeacher,
    NoisyStudent,
)

MAPPING_TO_MODEL = OrderedDict(
    [
        # mapping to ssl models
        ("PiModel", PiModel),
        ("TemporalEnsemble", TemporalEnsembleModel),
        ("CoTrain", CoTrain),
        ("TriTrain", TriTrain),
        ("MeanTeacher", MeanTeacher),
        ("NoisyStudent", NoisyStudent),
    ]
)

MAPPING_TO_TRAINER = OrderedDict(
    [
        # mapping to ssl trainer
        ("PiModel", TrainerWithUWScheduler),
        ("TemporalEnsemble", TrainerWithUWScheduler),
        ("CoTrain", TrainerForCoTraining),
        ("TriTrain", TrainerForTriTraining),
        ("MeanTeacher", TrainerWithUWScheduler),
        ("NoisyStudent", TrainerForNoisyStudent),
    ]
)

def generate_kwargs(hyperparam_dic):
    
    """
    Generator function for all combinations of hyperparameters from hyperparameter dictionary 
    
    Args:
    
    hyperparam_dic (:obj:`Dict` ): Hyperparameter dictionary.
    
    """
    if not hyperparam_dic:
        yield {}
        return
    
    numkeys = len(hyperparam_dic.keys())
    flatten_key_vals = ([[kys, vals] for kys, val_list in hyperparam_dic.items() for vals in val_list])

    for combs in itertools.combinations(np.arange(len(flatten_key_vals)), numkeys):
        
        kys=np.array(flatten_key_vals)[list(combs)][:,0]
        
        if len(set(kys))==len(kys):
            kwargs = {flatten_key_vals[i][0] : flatten_key_vals[i][1] for i in combs}
            yield kwargs
            
        else:
            continue

def check_and_replace(key, kwargs, args, basefunction):
    
    """
    Function to check for the "key" in "kwargs" or "args" and replace args 
    if key is found in args. In case it's not found either kwargs or args, 
    it takes the default option from basefunction. 
    
    Args:
    
    key (:obj:`str`): key to be searched.
    
    kwargs (:obj:`dict`): keyword argument dictionary to look through.
    
    args (:obj:`tuple`): arguments tuple to look through.
    
    basefunction(:obj: `train_with_ssl`): Base function around which wrapper 
    has been implemented.
    
    """

    if key in kwargs.keys():
        return kwargs.pop(key), args

    elif len(args):
        val = args[0]
        args_n = args[1:]
        del args
        gc.collect()
        return val, args_n
    
    else:
        signature = inspect.signature(basefunction)
        val = signature.parameters[key].default
        
        return val, args

def with_labeled_fraction(basefunction, labeled_fraction, *args, **kwargs):
    
    """
    Wrapper function around train_with_ssl implemented if a list of labeled_fractions is mentioned.
    
    Args:
    
    basefunction(:obj: `train_with_ssl`): Base function around which wrapper has been implemented.
    
    labeled_fraction (:obj:`list`): List of the labeled fraction of training data to be analysed. 
    This function uses ~dataset_utils.modify_datasets to divide the dataset into the fraction of 
    labeled dataset and unlabeled dataset. Then, each l_fr mentioned in labeled_fraction is analysed 
    seperately and results are stored in sup_stats and stats.
    
    kwargs: Remaining dictionary arguments for the train_with_ssl function.
    
    """

    sup_stats=[]
    stats=[]
    kwargs_defs = extract_keys(set_default_args, kwargs)
        
    dataset, args = check_and_replace('dataset', kwargs_defs, args, basefunction)
    model_name, args = check_and_replace('model_name', kwargs_defs, args, basefunction)
    ssl_model_type, args = check_and_replace('ssl_model_type', kwargs, args, basefunction)
    
    dataset, kwargs_ta = set_default_args(dataset, model_name, kwargs)
    
    batch_size = kwargs_ta['per_device_train_batch_size']
    
    for l_fr in labeled_fraction:
        
        dataset_new = modify_datasets(dataset.copy(), labeled_fr = l_fr, batchsize = batch_size,
                                      model_type = ssl_model_type)
        
        basefunction(dataset_new,
             model_name,
             ssl_model_type,
             *args, 
             sup_stats = sup_stats, 
             stats = stats, 
             l_fr=l_fr,
             **kwargs)
        
        del dataset_new
        gc.collect()
    
    return sup_stats, stats        

def wrapper_for_l_fr(func):
    
    @wraps(func)
    def decorator(*args, **kwargs):
                
        if 'labeled_fraction' in kwargs.keys() or (len(args) and isinstance(args[0], list)) :
            return with_labeled_fraction(func, *args, **kwargs)
        else:
            return func(*args, **kwargs)

    return decorator


@wrapper_for_l_fr
def train_with_ssl(dataset=None,
                    model_name="distilbert-base-uncased", 
                    ssl_model_type='PiModel',
                    text_column_name='sentence',
                    run_sup = False,
                    use_sup = False,
                    remove_dirs=True,
                    teacher_student_name = None,
                    num_labels = 2,
                    unsup_hp = {'w_ramprate':[0.001,0.01,0.1], 'alpha':[0.3,0.6,0.9]},
                    sup_stats = None,
                    stats = None,
                    l_fr=False,
                    **kwargs
                    ):
    
    """
    Function for training with semisupervised models during finetuning of the pretrained 
    transformer models. 
    
    Args:
    
    labeled_fraction (:obj:`list` ): Set up by wrapper_for_l_fr. List of labeled fraction 
    of training data to be analysed. In this case original dataset is divided into the 
    fraction of labeled dataset and unlabeled dataset.

    dataset (:obj:`~datasets.DatasetDict` ): Dataset dictionary containing labeled and 
    unlabeled data.
    
    model_name (:obj:`str` or :obj:`os.PathLike`): "pretrained_model_name_or_path" in 
    ~transformers.PreTrainedModel, please refer to its documentation for further information. 
    
    (i) In this case of a string, the `model id` of a pretrained model hosted inside a model 
    repo on huggingface.co.
    
    (ii) It could also be address of saved pretrained model.
    
    ssl_model_type (:obj:`str`): Semisupervised model type. 
    
    text_column_name (:obj:`str`): Column name for the text in the dataset.
    
    run_sup (:obj:`bool`): Whether to run a supervised model along with the semi supervised model
    for comparison.
    
    use_sup (:obj:`bool`): Whether to use the trained supervised as the starting point 
    the ssl model. If this is True, run_sup has to be true too. 
    
    remove_dirs (:obj:`bool`): Whether to remove dirs created by ~transformers.Trainer.

    teacher_student_name (:obj:`Tuple[`str`, `str`]): A Tuple for teacher and student name, 
    respectively for multi transformer models like MeanTeacher or NoisyStudent. Similar to model_name.
    
    num_labels (:obj:`int`): Total number of classes for the classification.
    
    unsup_hp (:obj:`dict`): The dictionary of all the hyperparameters in unsupervised part of model. 
    Check the documentation of ssl_model_type and the associated trainers before setting up this 
    dictionary. The train_with_ssl with then train the model on all the combinations of the 
    hyperparameters set in this dictionary.
    
    sup_stats (:obj:`list`): Used by the wrapper as an argument. List to save the supervised models 
    stats for comparison, if the run_sup is turned on. If labeled_fraction is not mentioned it is 
    created within this function.
    
    stats (:obj:`list`): Used by the wrapper as an argument. List to save the chosen semisupervised
    models (ssl_model_type) stats. Similar to sup_stats.
    
    l_fr(:obj:`float`): Used by the wrapper as an argument. Float value of fraction used as 
    labeled dataset.
    
    kwargs: Remaining dictionary arguments for the transformer.Trainer init function. Some of the 
    Trainer keyword are important for training like compute metrics, and tokenizer, 
    (see ~transformer.Trainer). If they are not mentioned, the default values would be picked,
    see default_args.set_default_args().
    
    Note: ~transformer.TrainingArgument which ~transformer.Trainer accepts as the args, here could 
    be given in same way. To distinguish arguments for the supervised trainer, it should be named 
    args_ta_sup for the supervised trainer and args_ta for the semisupervised trainer. There are 
    some default keys set in default_args file. If one just needs to change only some args and keep 
    rest of them as same default, then you can also set args_ta or args_ta_sup as a dictionary.
    
    For example setting:
    args_ta_sup = {learning_rate: 1} will only change learning rate to 1. Keep rest of them similar
    to what has been set in default_args. 
    
    Return:
    
    sup_stats (:obj:`list`): If used directly without the labeled_fraction and if run_sup is True 
    else returns empty list. Information of all the training history for supervised model.
    
    stats (:obj:`list`): If used directly without the labeled_fraction. Information of all the 
    training history for semi supervised model.
        
    """
    
    SSLModel = MAPPING_TO_MODEL[ssl_model_type]
        
    if not l_fr:
        
        sup_stats=[]
        stats=[]
        dataset, _ = set_default_args(dataset, model_name, kwargs)
    
    if use_sup: 
        run_sup=True
        if ssl_model_type in ['MeanTeacher', 'NoisyStudent']:
            raise TypeError( 'Supervised models cannot be used for multi semisupervised models starting points.' )
    
    args_ta_sup_in_keys = kwargs.pop('args_ta_sup', None)
    args_ta_in_keys = kwargs.pop('args_ta', None)
    
    args_ta_sup = get_default_ta_sup(f'runs/l_fr_{l_fr}') if args_ta_in_keys is None else args_ta_sup_in_keys.copy()
    
    output_dir_sup = args_ta_sup.output_dir
    
    if run_sup:
        
        model_sup = AutoModelForSequenceClassification.from_pretrained(model_name, num_labels = num_labels)
        trainer_sup = Trainer(
            model_sup,
            args_ta_sup,
            train_dataset=dataset["labeled"],
            eval_dataset=dataset["validation"],
            **kwargs
        )
        
        trainer_sup.train()
        best_model_checkpoint=trainer_sup.state.best_model_checkpoint
        trainer_sup.state.log_history.append(('l_fr', l_fr))
        sup_stats.append(trainer_sup.state.log_history)

        del trainer_sup, model_sup, args_ta_sup
        gc.collect()

    for kwargs_ssl in generate_kwargs(unsup_hp):
        logging_dir=''
        
        if l_fr: logging_dir += f'runs/l_fr_{l_fr}_'
        
        # For tensorboard.
        for k, v in kwargs_ssl.items(): logging_dir += k + '_' + str(v) + '_' 
        
        # For train_with_ssl plotter.
        kwargs_ssl_copy = kwargs_ssl.copy()
        
        # Seperate the keys for model and trainer
        kwargs_model = extract_keys(SSLModel, kwargs_ssl) 
        
        if use_sup == True:
            model_ssl = SSLModel(model_name=best_model_checkpoint, supervised_run=True,**kwargs_model)
            
        elif teacher_student_name is not None:
            model_ssl = SSLModel(teacher_student_name=teacher_student_name, **kwargs_model)
            
        else:
            model_ssl = SSLModel(model_name=model_name, **kwargs_model)
        
        args_ta = get_default_ta(logging_dir) if args_ta_in_keys is None else args_ta_in_keys.copy()
        
        output_dir = args_ta.output_dir

        trainer_ssl = MAPPING_TO_TRAINER[ssl_model_type](**kwargs_ssl, 
                                                            model=model_ssl,
                                                            args=args_ta,
                                                            dataset=dataset,
                                                            eval_dataset = dataset['validation'],
                                                            **kwargs
                                                            )
        trainer_ssl.train()
        
        
        for kys, vals in kwargs_ssl_copy.items(): trainer_ssl.state.log_history.append((kys, vals))
        
        if l_fr: trainer_ssl.state.log_history.append(('l_fr', l_fr))

        stats.append(trainer_ssl.state.log_history)
        
        del model_ssl, trainer_ssl, args_ta
        gc.collect()
    
    if remove_dirs:
        if os.path.exists(output_dir_sup): 
            shutil.rmtree(output_dir_sup)
            
        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)
            
    if not l_fr: 
        return sup_stats, stats



import torch
from transformers import Trainer
from torch.utils.data.dataloader import DataLoader
from torch.utils.data.sampler import RandomSampler, SequentialSampler
from transformers import AutoTokenizer
from transformers import AutoModelForSequenceClassification, TrainingArguments, Trainer
from .dataset_utils import modify_datasets, SimpleDataset, extract_keys
import shutil
import numpy as np
import pandas as pd
import gc
from collections import defaultdict
import itertools
import inspect
from transformers.data.data_collator import DataCollatorWithPadding
from transformers.optimization import get_scheduler, AdamW
from torch.optim.lr_scheduler import MultiplicativeLR
from copy import deepcopy
from typing import Optional, Tuple
from torch.optim import Optimizer
from transformers.trainer_callback import TrainerState
from transformers.trainer_pt_utils import nested_detach

def get_linear_schedule_with_minlr(optimizer: Optimizer, num_warmup_steps: int, num_training_steps: int , last_epoch: int = -1, min_lr : int = 1e-07):
    
    """
    Creates a scheduler with a learning rate that linearly decreases but saturates at min_lr value. 
    
    Args:
        optimizer (:class:`~torch.optim.Optimizer`):
            The optimizer for which to schedule the learning rate.
        num_warmup_steps (:obj:`int`):
            The number of steps for the warmup phase.
        num_training_steps (:obj:`int`):
            The total number of training steps.
        num_cycles (:obj:`int`, `optional`, defaults to 1):
            The number of hard restarts to use.
        last_epoch (:obj:`int`, `optional`, defaults to -1):
            The index of the last epoch when resuming training.
        min_lr (:obj:`int`, `optional`, defaults to 1e-07):
            The value of minimum learning rate where it should saturate.

    Return:
        :obj:`torch.optim.lr_scheduler.MultiplicativeLR` with the appropriate schedule.
    """

    init_lr = optimizer.defaults['lr']
    
    def lr_lambda(current_step: int):
        
        steps_done = float(num_training_steps - current_step)
        
        if current_step>1:
            mul_fac = steps_done / max(steps_done+1, 1)
        else:
            mul_fac = steps_done /(num_training_steps)
        
        if mul_fac*init_lr>min_lr:
            return mul_fac  
        else:
            return 1

    return MultiplicativeLR(optimizer, lr_lambda, last_epoch)

class RemoveUnusedColumnMixing:
    
    def _remove_unused_columns(self, dataset: "datasets.Dataset", description: Optional[str] = None):
        
        """
        Simply replacing model.pretrained_model.forward and model.teacher.forward in the 
        case of MeanTeacher and NoisyStudent from model.forward in
        ~transformers.Trainer._remove_unused_columns.
        """        

        if not self.args.remove_unused_columns:
            return
        # Inspect model forward signature to keep only the arguments it accepts.
        if self.model.type_=='MeanTeacher' or self.model.type_=='NoisyStudent':
            signature = inspect.signature(self.model.teacher.forward)
        else:
            signature = inspect.signature(self.model.pretrained_model.forward)

        signature_columns = list(signature.parameters.keys())
        # Labels may be named label or label_ids, the default data collator handles that.
        signature_columns += ["label", "label_ids"]
        columns = [k for k in signature_columns if k in dataset.column_names]

        dataset.set_format(type=dataset.format["type"], columns=columns)


class UWScheduler:
    
    """
    Unsupervised weights scheduler for changing the unsupervised weight
    of the semi supervised learning models. Also contains methods any 
    other kinds variables updates required by the models. For example, 
    PiModel, TemporalEnsembling model, and Mean Teacher. In this implementation, 
    it's  based on the composition with learning scheduler from pytorch and it 
    works best ~transformers.Trainer without having to rewrite train method. 
    # TODO: cleaner version with the rewritten Trainer.train method.
    
    Args:
        
    lr_scheduler (:obj:`torch.optim.lr_scheduler`): Learning scheduler object. 
    
    trainer (:obj:`~TrainerWithUWScheduler`): Trainer object.
    
    unsup_start_epochs (:obj:`int`): value of epoch at which the unsupervised 
    weights should start updating.

    max_w (:obj:`float`): maximum value of weight that the unsup_weight from model 
    could reach.

    update_teacher_steps (:obj:`int`): useful for MeanTeacher, sets the interval after 
    which teacher variables should be updated.
    
    w_ramprate (:obj:`float`): linear rate at which the unsupervised weight would be 
    increased from the initial value.
    
    update_weights_steps (:obj:`int`): interval steps after which unsupervised weight 
    would be updated by the w_ramprate.
    
    Class attributes:
     
    -**step_in_epochs**: Number of steps (batch passes) in an epoch.

    -**local_step**: keeps track of the times unsupervised weight has been changed.
     
    """

    def __init__(self, lr_scheduler, trainer, unsup_start_epochs=0, max_w=1,
                 update_teacher_steps=False, w_ramprate=1, update_weights_steps=1):
        
        self.trainer = trainer
        self.lr_scheduler = lr_scheduler
        self.local_step=0
        self.max_w=max_w
        self.unsup_start_epochs = unsup_start_epochs
        self.update_weights_steps = update_weights_steps
        self.w_ramprate = w_ramprate
        self.steps_in_epoch = len(self.trainer.train_dataset)//self.trainer.args.train_batch_size
        self.update_teacher_steps = update_memory_logits if update_teacher_steps else self.steps_in_epoch

        
    def step(self):
        
        """
        Implementation of composition of the pytorch learning rate scheduler step function
        with schedule of unsupervised weights. Also implements updating the memory logits 
        for TemporalEnsembleModel and updating teacher variables for MeanTeacher model.
        
        """

        self.lr_scheduler.step()

        if self.trainer.state.epoch > self.unsup_start_epochs  and self.is_true(self.update_weights_steps):
            self.trainer.model.unsup_weight = min(self.max_w, self.w_ramprate*self.local_step)
            self.local_step += 1
            
        if self.trainer.model.type_=="TemporalEnsembleModel" and self.is_true(self.steps_in_epoch):
            self.trainer.model.update_memory_logits(int(self.trainer.state.epoch + 1))
        
        if self.trainer.model.type_=="MeanTeacher" and self.is_true(self.update_teacher_steps):
            self.trainer.model.update_teacher_variables() 
   
    def is_true(self, value):
        """
        A simple checker function to if it is time to update things depending on the value of value.
        """
        return (self.trainer.state.global_step and (self.trainer.state.global_step + 1)%value==0)
        
    def __getattr__(self, name):
        """
        Needed for the calls from ~transformers.Trainer.train() method.
        """
        return getattr(self.lr_scheduler, name)



class TrainerWithUWScheduler(RemoveUnusedColumnMixing, Trainer):
    
    """
    Subclass of ~transformers.Trainer with minimal code change and integration 
    with unsupervised weight scheduler. 
    
    Args:
        
    kwargs_uw: dictionary of arguments to be used by UWScheduler. 
    
    kwargs: dictionary arguments for the ~transformers.Trainer, of dataset used 
    by the trainer and could also include arguments of UWScheduler.
    
    Note: dataset for training can be given to the trainer in two ways.
    
    (i) dataset: In this case, it should the naming scheme of dataset_utils.modify_datasets. 
    
    (ii)train_dataset: Same naming scheme as used by ~transformers.Trainer. 
         
    """

    def __init__(self, kwargs_uw=None, *args, **kwargs):
        
        self.kwargs_uw = kwargs_uw if kwargs_uw else extract_keys(UWScheduler, kwargs)
        
        dataset = kwargs.pop('dataset', None)
        
        if dataset:
            kwargs['train_dataset'] = dataset["train"]        
        
        super().__init__(*args, **kwargs)  

        self.check_for_consistency()
        
    def create_optimizer_and_scheduler(self, num_training_steps):
        
        """
        Overriden ~transformers.Trainer.create_optimizer_and_scheduler with integration
        with the UWScheduler to its Trainer.lr_scheduler object
        
        """        
        super().create_optimizer_and_scheduler(num_training_steps)
        
        self.lr_scheduler = UWScheduler(self.lr_scheduler, self, **self.kwargs_uw)

    def get_train_dataloader(self):

        """
        Slightly changed ~transformers.Trainer.get_train_dataloader as models used in 
        Trainer do not allow for mixing of labeled and unlabeled data. So changing to 
        SequentialSampler instead of RandomSampler. 
        
        """        
        
        train_sampler = SequentialSampler(self.train_dataset)
        
        return DataLoader(
            self.train_dataset,
            batch_size=self.args.train_batch_size,
            sampler=train_sampler,
            collate_fn=self.data_collator,
            drop_last=self.args.dataloader_drop_last,
            num_workers=self.args.dataloader_num_workers,
        )
    
    def check_for_consistency(self):
        """
        Checks if the labeled and unlabeled are present in the same minibatch, raises error if they are. 
        """        
        
        labels = torch.tensor(self.train_dataset['label'])
        batch_size = self.args.per_device_train_batch_size
        
        for ind in range(len(labels)//batch_size):
            
            min_batch = labels[ind*batch_size : (ind+1)*batch_size]
            if not (all(min_batch>=0) or all(min_batch<0)):
                raise RuntimeError('Mixing of labeled and unlabeled examples is not allowed.')    
 
        
class BaseForMMTrainer(RemoveUnusedColumnMixing, Trainer):
    
    """
    Base class for all the mutimodel trainers. This class contains 
    the methods which are used by the Trainers which helps in training 
    the semi supervised way which have mutiple models .

    """
    def get_dataloader(self, dataset, sequential=False):
        
        """
        Slightly changed ~transformers.Trainer.get_train_dataloader
        with the flexibility to change between sequential and RandomSampler. 
        """        

        if sequential==True:
            sampler = SequentialSampler(dataset)
        else:
            sampler = RandomSampler(dataset)

        return DataLoader(
            dataset,
            batch_size=self.args.train_batch_size,
            sampler = sampler,
            collate_fn=self.data_collator,
            drop_last=self.args.dataloader_drop_last,
            num_workers=self.args.dataloader_num_workers,
        )

    
    def pre_train_init(self, num_training_steps):
        
        """
        Collection of all the callback functions called before initializing 
        the training. See ~transformers.Trainer.train() for more details.
        
        Args:
        num_training_steps (:obj:`int`): total number of training which are 
        calculated by number of mini batches per epoch * number of epochs.

        """        
       
        self.state = TrainerState()
        self.create_optimizer_and_scheduler(num_training_steps)
        
        if self.use_min_lr_scheduler is not None:
            self.lr_scheduler = get_linear_schedule_with_minlr(self.optimizer, num_warmup_steps=self.args.warmup_steps,
                                                          num_training_steps=num_training_steps, min_lr=self.min_lr)
        self.callback_handler.model = self.model
        self.callback_handler.optimizer = self.optimizer
        self.callback_handler.lr_scheduler = self.lr_scheduler
        self.control = self.callback_handler.on_train_begin(self.args, self.state, self.control)
        self._total_loss_scalar = 0.0
        self._globalstep_last_logged = self.state.global_step
                    
    def post_epoch(self, step, epoch, tr_loss):
        
        """
        Collection of all the callback functions called 
        after the epoch is done. See ~transformers.Trainer.train() for more details.
        
        Args:
        step (:obj:`int`): step number, number of steps passed out of 
        num_training_steps used in pre_train_init method.
        
        epoch (:obj:`int`): epoch passed.
        
        tr_loss(:obj:`float`): training loss.

        """        
        
        torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.args.max_grad_norm)                
        self.optimizer.step()
        self.lr_scheduler.step()
        self.model.zero_grad()
        self.state.global_step += 1
        self.state.epoch = self.global_epoch + epoch + (step + 1) / self.steps_in_epoch
        self.control = self.callback_handler.on_step_end(self.args, self.state, self.control)
        self._maybe_log_save_evaluate(tr_loss, self.model, trial=None, epoch=epoch)

    def equate_lengths(self, model1_train, model2_train):
        
        """
        A method useful for CoTrain, TriTrain model training. It finds 
        whichever model dataset has less training examples and then 
        equates them using SimpleDataset.extend_length method

        Args:
        model1_train (:obj:`SimpleDataset`): Training dataset for model 1. 
        
        model2_train (:obj:`SimpleDataset`): Training dataset for model 2.

        """        

        if len(model1_train) > len(model2_train):
            model2_train.extend_length(len(model1_train))
        elif len(model2_train) > len(model1_train):
            model1_train.extend_length(len(model2_train))
            
    def confi_prediction(self, logits_m1, logits_m2, logits_m3=None):
        
        """
        Prediction made based on between confidence of the models. First 
        checks whichever model has the highest probability (confidence) 
        on a given example and choses that class as the final answer. 
        
        Args:
        logits_m1 (:obj:`torch.FloatTensor`): logits recieved from model 1. 

        logits_m2 (:obj:`torch.FloatTensor`): logits from model 2.

        logits_m3 (:obj:`torch.FloatTensor`): logits from model 3.

        """        
        
        p_label1, labels1 = torch.max(logits_m1, 1)
        p_label2, labels2 = torch.max(logits_m2, 1)
        
        logits_out = torch.zeros(logits_m1.size(), device=logits_m1.device)
                
        if logits_m3 is None:

            m1_confi = p_label1>=p_label2
            m2_confi = p_label2>=p_label1

        else:
            p_label3, labels3 = torch.max(logits_m3, 1)
            m1_confi = torch.logical_and(p_label1>=p_label2, p_label1>=p_label3)
            m2_confi = torch.logical_and(p_label2>=p_label1, p_label2>=p_label3)
            m3_confi = torch.logical_and(p_label3>=p_label1, p_label3>=p_label2)
            logits_out[m3_confi, :] = logits_m3[m3_confi, :]

        logits_out[m1_confi, :] = logits_m1[m1_confi, :]
        logits_out[m2_confi, :] = logits_m2[m2_confi, :]
        
        return logits_out
           
    def prediction_step(
        self,
        model,
        inputs,
        prediction_loss_only,
        ignore_keys= None,
    ):
        
        """
        Slightly changed ~transformers.Trainer.prediction_step using 
        confi_prediction method to find prediction of Cotrain and TriTrain Models.
        Used during evaluation step.
        """
        has_labels = all(inputs.get(k) is not None for k in self.label_names)
        inputs = self._prepare_inputs(inputs)
        if ignore_keys is None:
            if hasattr(self.model, "config"):
                ignore_keys = getattr(self.model.config, "keys_to_ignore_at_inference", [])
            else:
                ignore_keys = []

        with torch.no_grad():
            if self.use_amp:
                with autocast():
                    outputs = model(**inputs)
            else:
                outputs = model(**inputs)
            if has_labels:
                if self.label_smoother is not None and "labels" in inputs:
                    loss = self.label_smoother(outputs, inputs["labels"]).mean().detach()
                else:
                    loss = (outputs["loss"] if isinstance(outputs, dict) else outputs[0]).mean().detach()
                if isinstance(outputs, dict):
                    logits = {k:v for k, v in outputs.items() if k not in ignore_keys + ["loss"]}
                    logits = self.confi_prediction(**logits)
                else:
                    
                    logits = outputs[1:]
            else:
                loss = None
                if isinstance(outputs, dict):
                    logits = tuple(v for k, v in outputs.items() if k not in ignore_keys)
                else:
                    logits = outputs
            if self.args.past_index >= 0:
                self._past = outputs[self.args.past_index if has_labels else self.args.past_index - 1]

        if prediction_loss_only:
            return (loss, None, None)

        logits = nested_detach(logits)
        if len(logits) == 1:
            logits = logits[0]

        if has_labels:
            labels = nested_detach(tuple(inputs.get(name) for name in self.label_names))
            if len(labels) == 1:
                labels = labels[0]
        else:
            labels = None

        return (loss, logits, labels)
    
        

class TrainerForCoTraining(BaseForMMTrainer):
    
    """
    Subclass of ~transformers.Trainer with changes for Cotrain Model.

    Args:
    
    epoch_per_cotrain (:obj:`int`, `optional`, defaults to 2): Number of epochs to 
    pass through training data while going through one iteration of cotraining.
    
    exchange_threshold (:obj:`int`, `optional`, defaults to 20): Threshold value 
    of exchange between the model, below which co training is stopped.

    ntimes_before_saturation (:obj:`int`, `optional`, defaults to 2): In the case of 
    get_linear_schedule_with_minlr, number of times should go over all model dataset 
    + unlabeled dataset before saturating. 
    
    p_threshold (:obj:`float`): threshold probability for considering exchange between models.
    
    use_min_lr_schedule (:obj:`bool`): Whether to use linear_schedule_with_minlr.
    
    min_lr (:obj:`int`, `optional`, defaults to 1e-07): The value of minimum learning rate 
    used by linear_schedule_with_minlr.
    
    show_exchange (:obj:`bool`): Whether to print the exchange happening between models.
    
    max_passes (:obj:`int`): Maximum number of passes through the all the datasets.


    Note: dataset for training can be given to the trainer in two ways.
    
    (i) dataset: In this case, it should the naming scheme of dataset_utils.modify_datasets. 
    
    (ii)dataset_model1, dataset_model2, unlabeled: dataset for model1, model2 and 
    unlabeled dataset respectively.
    
    
    Class attributes:
     
    -**total_dataset_len**: total length of all datasets including model1, model2 and unlabeled.

    -**global_epoch**: number of epoch globally including the epochs over 
    model1_dataset+unlabeled and model2_dataset +unlabeled.
         
    """
 
    def __init__(self,  epoch_per_cotrain=2, exchange_threshold=20, ntimes_before_saturation=2,  p_threshold=0.65, use_min_lr_scheduler=True, min_lr=1e-07, show_exchange=True, max_passes=1, *args, **kwargs):
        
        dataset = kwargs.pop('dataset', None)
        
        self.dataset_model1 = dataset['labeled1'] if dataset else kwargs.pop("dataset_model1")
        self.dataset_model2 = dataset['labeled2'] if dataset else kwargs.pop("dataset_model2")
        self.dataset_unlabeled = dataset['unlabeled'] if dataset else kwargs.pop("unlabeled")
        self.use_min_lr_scheduler=use_min_lr_scheduler
        ##needed ~transformers.Trainer's constructor to not to initate the lr_scheduler. 
        self.lr_scheduler = use_min_lr_scheduler
        
        super().__init__(*args, **kwargs)
        
        self.min_lr = min_lr
        self.epoch_per_cotrain = epoch_per_cotrain
        self.exchange_threshold = exchange_threshold
        self.ntimes_before_saturation = ntimes_before_saturation
        self.p_threshold = p_threshold
        self.show_exchange = show_exchange
        self.max_passes = max_passes
        
        self._remove_unused_columns(self.dataset_unlabeled, description="unlabeled")
        self._remove_unused_columns(self.dataset_model1, description="labeled1")
        self._remove_unused_columns(self.dataset_model2, description="labeled2")
        self.dataset_model1 = SimpleDataset(self.dataset_model1)
        self.dataset_model2 = SimpleDataset(self.dataset_model2)
        self.dataset_unlabeled = SimpleDataset(self.dataset_unlabeled)
        
        self.total_dataset_len = (len(self.dataset_model1) + 
                                  len(self.dataset_model2) +
                                  len(self.dataset_unlabeled))
        self.global_epoch=0
        

    def train(self):  
        """
        Train method for CoTrainer. Performs exchange of unlabeled dataset 
        between model1 and model2 until exchange condition is true, see 
        exchange_unlabeled_data(). 
        """
        
        num_training_steps = int(self.ntimes_before_saturation * self.total_dataset_len * 
                                 self.epoch_per_cotrain/self.args.train_batch_size)        
        
        self.pre_train_init(num_training_steps)
        self.cotrain(self.dataset_model1, self.dataset_model2)
        
        exchange=True
        count_passes=0
        while exchange==True and count_passes<self.max_passes:
            ul_dataloader = self.get_dataloader(self.dataset_unlabeled)
            with torch.no_grad():
                exchange = self.exchange_unlabeled_data(ul_dataloader)
            
            if exchange : self.cotrain(self.dataset_model1, self.dataset_model2)

            count_passes += 1
            

    def cotrain(self, model1_train, model2_train):
        """
        Method where the real co training takes place. Firstly, the dataset 
        for the both models are equated with randomly repeated examples from 
        training dataset, see SimpleDataset.extend_length(). Then, a batch from 
        both model1 and model2 dataset is passed to cotrain_forward, where applies 
        the orthogonality based between the classifier layers and calculates the 
        total loss from both the batches.
        
        Args:
        model1_train (:obj:`SimpleDataset`): Training dataset for model 1 after
        the exchange_unlabeled_data has been used. 
        
        model2_train (:obj:`SimpleDataset`): Training dataset for model 2 after 
        the exchange_unlabeled_data has been used.

        """

        self.equate_lengths(model1_train, model2_train)
        model1_dataloader = self.get_dataloader(model1_train)
        model2_dataloader = self.get_dataloader(model2_train)            
        self.steps_in_epoch = len(model1_dataloader)
        
        self.model.zero_grad()
        
        for epoch in range(self.epoch_per_cotrain):
            
            self.control = self.callback_handler.on_epoch_begin(self.args, self.state, self.control)
            
            for step, (batch_for_model1, batch_for_model2) in enumerate(zip(model1_dataloader, model2_dataloader)):
                
                self.control = self.callback_handler.on_step_begin(self.args, self.state, self.control)

                self.model.train()
                batch_for_model1 = self._prepare_inputs(batch_for_model1)
                batch_for_model2 = self._prepare_inputs(batch_for_model2)
                
                outputs = self.model.cotrain_forward(batch_for_model1, batch_for_model2)
                tr_loss = outputs.loss
                tr_loss.backward()
                self.post_epoch(step, epoch, tr_loss)
                
            self.control = self.callback_handler.on_epoch_end(self.args, self.state, self.control)

            if self.control.should_evaluate: self.evaluate()
            
        self.global_epoch += self.epoch_per_cotrain    
    
            
    def exchange_unlabeled_data(self, ul_dataloader):
        
        """
        Method to exchange the unlabeled dataset between models. Examples on which
        model1 is confident on(above p_threshold) are given to model2 to train on and vice versa. 
        
        Args:
        ul_dataloader (class:`~torch.utils.data.DataLoader`): dataloader for unlabeled dataset.
        
        Returns:
        exchange (:obj:`bool`): If the number of the examples exchanged are above the 
        exchange_threshold.
        
        dataset_model1, dataset_model2 (:obj:`SimpleDataset`): dataset for model1 and 
        model2 with added unlabeled data. 
        
        """
        
        self.dataset_model1.reset()
        self.dataset_model2.reset()
        added = 0
        
        for batch_index, inputs in enumerate(ul_dataloader):
            
            inputs = self._prepare_inputs(inputs)

            logits1 = self.model.simple_forward(**inputs, classifier_num=0)
            logits2 = self.model.simple_forward(**inputs, classifier_num=1)
            
            p_label1, labels1 = torch.max(logits1, 1)
            p_label2, labels2 = torch.max(logits2, 1)
            
            ##labels on which model 1 is confident on.
            labels_m1_confi = torch.logical_and(p_label1>=self.p_threshold, p_label2<self.p_threshold)
            labels_m1_confi = torch.logical_and(labels_m1_confi, labels1!=labels2)
            
            ##labels on which model 2 is confident on.
            labels_m2_confi = torch.logical_and(p_label2>=self.p_threshold, p_label1<self.p_threshold)
            labels_m2_confi = torch.logical_and(labels_m2_confi, labels1!=labels2)
            
            inputs['label'] = labels2
            added += self.dataset_model1.append(inputs, labels_m2_confi, batch_index)
            
            inputs['label'] = labels1
            added += self.dataset_model2.append(inputs, labels_m1_confi, batch_index)
        
        exchange = True if added>self.exchange_threshold else False

        if self.show_exchange: print(f'Number of unlabeled examples exchanged: {added.item()}')

        self.dataset_model1 = self.dataset_model1.reformat()
        self.dataset_model2 = self.dataset_model2.reformat()

        return exchange
        

class TrainerForTriTraining(TrainerForCoTraining):
    
    """
    Subclass of ~TrainerForCoTraining with adding of a third model for Tritrain Model.

    Args:
    
    procedure (:obj:`str`, `optional`, defaults to 'agreement'): Whether 
    to train with TriTraining with agreement <https://ieeexplore.ieee.org/document/1512038> 
    or with disagreement <https://www.aclweb.org/anthology/P10-2038/>.

    Note: dataset for training can be given to Trainer as same way as TrainerForCoTraining 
    but with and addition of the third dataset for model3.
    
    (i)dataset: In this case, it should the naming scheme of dataset_utils.modify_datasets. 
    
    (ii)dataset_model3: dataset for model3.
    
    """

    def __init__(self, procedure='agreement', epoch_per_tritrain=2, *args, **kwargs):
        
        
        dataset_bool = True if 'dataset' in kwargs.keys() else False
        
        self.dataset_model3 = kwargs['dataset']['labeled3'] if dataset_bool else kwargs.pop("dataset_model3")
        
        super().__init__(*args, **kwargs)
        
        if procedure=='agreement':
            self.exchange_proc = self.agreement_proc
        else:
            self.exchange_proc = self.disagreement_proc

        self.epoch_per_tritrain = epoch_per_tritrain

        self.total_dataset_len = (len(self.dataset_model1) + 
                                  len(self.dataset_model2) + 
                                  len(self.dataset_model3) + 
                                  len(self.dataset_unlabeled))
    
        self._remove_unused_columns(self.dataset_model3, description="labeled3")
        self.dataset_model3 = SimpleDataset(self.dataset_model3)
        
    def train(self):
        """
        Train method for TriTrainer. Performs exchange of unlabeled dataset between
        three models, model1, model2 and model3 until exchange condition is true, 
        see exchange_unlabeled_data(). 
        """
        
        num_training_steps = int(self.ntimes_before_saturation * self.total_dataset_len * 
                                 self.epoch_per_cotrain/self.args.train_batch_size)
        
        
        self.pre_train_init(num_training_steps)
        
        self.tri_train(self.dataset_model1, self.dataset_model2, self.dataset_model3)

        exchange=True 
        count_passes=0
        while exchange==True or count_passes<self.max_passes:
            ul_dataloader = self.get_dataloader(self.dataset_unlabeled, True)
            
            with torch.no_grad():
                exchange = self.exchange_unlabeled_data(ul_dataloader)
                
            if exchange : self.tri_train(self.dataset_model1, self.dataset_model2, self.dataset_model3)
            count_passes += 1


    def tri_train(self, model1_train, model2_train, model3_train):
        
        """
        Method for tri training. Same procedure as TrainerForCoTraining.co_train 
        but with the addition of training for the third model.
        """        
        
        self.equate_lengths(model1_train, model2_train)
        self.equate_lengths(model2_train, model3_train)
        
        self.model.zero_grad()
        
        model1_dataloader = self.get_dataloader(model1_train)
        model2_dataloader = self.get_dataloader(model2_train)
        model3_dataloader = self.get_dataloader(model3_train)
        self.steps_in_epoch = len(model3_dataloader)
        
        for epoch in range(self.epoch_per_tritrain):            
            
            self.control = self.callback_handler.on_epoch_begin(self.args, self.state, self.control)
            
            for step, batches in enumerate(zip(model1_dataloader, model2_dataloader, model3_dataloader)):
                
                self.control = self.callback_handler.on_step_begin(self.args, self.state, self.control)

                self.model.train()

                batch_for_model1 = self._prepare_inputs(batches[0])                
                batch_for_model2 = self._prepare_inputs(batches[1])                
                batch_for_model3 = self._prepare_inputs(batches[2])
                
                ct_outputs = self.model.cotrain_forward(batch_for_model1, batch_for_model2)
                m3_output = self.model.m3_forward(**batch_for_model3)
                tr_loss = ct_outputs.loss + m3_output.loss
                tr_loss.backward()
                
                self.post_epoch(step, epoch, tr_loss)
                
            self.control = self.callback_handler.on_epoch_end(self.args, self.state, self.control)
            
            if self.control.should_evaluate: self.evaluate()
        
        self.global_epoch += self.epoch_per_tritrain
    
    def disagreement_proc(self, labels_confi, la, lb, l_compare):
        """
        Function to implement disagreement procedure during the exchange of unlabeled data.
        
        """        
        labels_confi = torch.logical_and(labels_confi, la==lb)
        labels_confi = torch.logical_and(labels_confi, la!=l_compare)
        return torch.logical_and(labels_confi, lb!=l_compare)
            
    def agreement_proc(self, labels_confi, la, lb, l_compare):        
        """
        Function to implement agreement procedure during the exchange of unlabeled data.
        
        """        
        return torch.logical_and(labels_confi, la==lb)
    
    
    def exchange_unlabeled_data(self, ul_dataloader):
        
        """
        Method to exchange the unlabeled dataset between models. Based on 
        chosen exchange procedure. Rest is similar to TrainerForCoTraining.exchange_unlabeled_data().
        
        Args:
        ul_dataloader (class:`~torch.utils.data.DataLoader`): dataloader for unlabeled dataset.
        
        Returns:
        exchange (:obj:`bool`): If the number of the examples exchanged are above the exchange_threshold.
        """
 
        self.dataset_model1.reset()
        self.dataset_model2.reset()
        self.dataset_model3.reset()

        added=0
        for batch_index, inputs in enumerate(ul_dataloader):
            
            inputs = self._prepare_inputs(inputs)

            logits1 = self.model.simple_forward(**inputs, classifier_num=0)
            logits2 = self.model.simple_forward(**inputs, classifier_num=1)
            logits3 = self.model.simple_forward(**inputs, classifier_num=2)
            
            p_label1, labels1 = torch.max(logits1, 1)
            p_label2, labels2 = torch.max(logits2, 1)
            p_label3, labels3 = torch.max(logits3, 1)
            
            labels_m1m2_confi = torch.logical_and(p_label1>=self.p_threshold, p_label2>=self.p_threshold)
            labels_m1m2_confi = self.exchange_proc(labels_m1m2_confi, labels1, labels2, labels3)
            
            labels_m2m3_confi = torch.logical_and(p_label2>=self.p_threshold, p_label3>=self.p_threshold)
            labels_m2m3_confi = self.exchange_proc(labels_m2m3_confi, labels2, labels3, labels1)

            labels_m1m3_confi = torch.logical_and(p_label1>=self.p_threshold, p_label3>=self.p_threshold)
            labels_m1m3_confi = self.exchange_proc(labels_m1m3_confi, labels1, labels3, labels2)

            inputs['label'] = labels2
            added += self.dataset_model1.append(inputs, labels_m2m3_confi, batch_index)

            inputs['label'] = labels1
            added += self.dataset_model2.append(inputs, labels_m1m3_confi, batch_index)

            inputs['label'] = labels2
            added += self.dataset_model3.append(inputs, labels_m1m2_confi, batch_index)

        exchange = True if added>self.exchange_threshold else False
        
        if self.show_exchange: print(f'Number of unlabeled examples exchanged: {added.item()}')
        
        self.dataset_model1 = self.dataset_model1.reformat()
        self.dataset_model2 = self.dataset_model2.reformat()
        self.dataset_model3 = self.dataset_model3.reformat()
        
        return exchange
    
    
class TrainerForNoisyStudent(BaseForMMTrainer):
    
    """
    Subclass of ~transformers.Trainer for the noisy student model.

    Args:
    
    min_lr (:obj:`int`, `optional`, defaults to 1e-07): The value of minimum 
    learning rate used by get_linear_schedule_with_minlr.
    
    epoch_per_ts_iter (:obj:`int`, `optional`, defaults to 1): Number of epochs 
    to pass during each teacher student iteration.
    
    ts_iter (:obj:`int`, `optional`, defaults to 3): Number of teacher student 
    iterations during training, in which student is again used as the teacher. 

    ntimes_before_saturation (:obj:`int`, `optional`, defaults to 2): In the case 
    of get_linear_schedule_with_minlr, number of times should go over all model 
    dataset + unlabeled dataset before saturating. 

    Note: dataset for training can be given to Trainer in two ways.
    
    (i)dataset: In this case, it should the naming scheme of dataset_utils.modify_datasets. 
    
    (ii)dataset_labeled, dataset_unlabeled: dataset for labeled and unlabeled data.
    
    """

    def __init__(self, min_lr=1e-07, epoch_per_ts_iter=1, ts_iter=3, ntimes_before_saturation=2, reduce_init_lr_factor=1, use_min_lr_scheduler=None, *args, **kwargs):
        
        dataset = kwargs.pop('dataset', None)
        
        self.use_min_lr_scheduler = use_min_lr_scheduler
        self.dataset_labeled = dataset['labeled'] if dataset else kwargs.pop("dataset_labeled")
        self.dataset_unlabeled = dataset['unlabeled'] if dataset else kwargs.pop("dataset_unlabeled")
        self.lr_scheduler = use_min_lr_scheduler
        
        super().__init__(*args, **kwargs)
        
        self._remove_unused_columns(self.dataset_unlabeled, description="unlabeled")
        self._remove_unused_columns(self.dataset_labeled, description="labeled")
        self.dataset_labeled = SimpleDataset(self.dataset_labeled)
        self.dataset_unlabeled = SimpleDataset(self.dataset_unlabeled)
        
        self.min_lr = min_lr
        self.epoch_per_ts_iter = epoch_per_ts_iter
        self.ts_iter = ts_iter
        self.ntimes_before_saturation = ntimes_before_saturation
        self.reduce_init_lr_factor = reduce_init_lr_factor
        self.total_dataset_len = (len(self.dataset_labeled) + 
                                  len(self.dataset_unlabeled))

        self.global_epoch=0
        ##repoint to transformers.Trainer.prediction_step() method for this case.
        self.prediction_step = super(BaseForMMTrainer, self).prediction_step
    
    def train(self):
        """
        Train method for Noisy student. Trains the teacher and student 
        with exchange (replacing teacher with the student) at the end of 
        every iteration. In this case both optimizer and learning rate scheduler 
        at reinitiated at the end of an iteration. 
        
        """
        
        for _ in range(self.ts_iter):
            self.dataset_labeled.reset()
            self.train_and_reset(self.model.teacher)
            
            self.psuedo_label()
            
            self.train_and_reset(self.model.student)
            self.exchange_models()
            self.args.learning_rate /= self.reduce_init_lr_factor


    def train_and_reset(self, model):
        """
        Trains either teacher or student then resets the training 
        variables like optimizer and scheduler. 
        """

        self.pre_train_init(self.num_training_steps_)
        self.train_one_model(model)
        
        del self.optimizer, self.lr_scheduler
        self.lr_scheduler = self.use_min_lr_scheduler
        self.optimizer = None

    @property
    def num_training_steps_(self):
        return int(len(self.dataset_labeled)*self.epoch_per_ts_iter/self.args.train_batch_size)

    def train_one_model(self, model):        
        """
        Common method for training either the teacher or student model.
        """

        train_dataloader = self.get_dataloader(self.dataset_labeled)
        self.model.zero_grad()
        self.steps_in_epoch = len(train_dataloader)
        
        for epoch in range(self.epoch_per_ts_iter):            
            
            self.control = self.callback_handler.on_epoch_begin(self.args, self.state, self.control)
            
            for step, inputs in enumerate(train_dataloader):
                
                self.control = self.callback_handler.on_step_begin(self.args, self.state, self.control)
                self.model.train()
                
                inputs = self._prepare_inputs(inputs)                
                tr_loss = self.model(**inputs).loss
                tr_loss.backward()
                
                self.post_epoch(step, epoch, tr_loss)
                
            self.control = self.callback_handler.on_epoch_end(self.args, self.state, self.control)
            if self.control.should_evaluate: self.evaluate()

            
        self.global_epoch += self.epoch_per_ts_iter
            


    def psuedo_label(self):        
        """
        Method for generating psuedo_label for by teacher model for student.
        """
        
        ul_dataloader = self.get_dataloader(self.dataset_unlabeled)
        
        for inputs in ul_dataloader:
            
            inputs = self._prepare_inputs(inputs)
            logits = self.model.teacher(**inputs).logits
            _, labels = torch.max(logits, 1)
            inputs['label'] = labels
            self.dataset_labeled.append(inputs)            

        self.dataset_labeled = self.dataset_labeled.reformat()
            
    def exchange_models(self):
        """
        Method for changing student model into teacher model. 
        """
        
        del self.model.teacher
        gc.collect()
        
        self.model.teacher = deepcopy(self.model.student)
        self.model.teacher.dropout.p = self.model.teacher_dropout



import pandas as pd
import matplotlib.pyplot as plt
import matplotlib as mpl
import gc
from .training_args import generate_kwargs

def add_end_args(from_fn):
    
    addFromstr = f'Adding Args from function-> {from_fn.__name__}'
    def docstring_decorator(fn):
        index = from_fn.__doc__.find('Args:')
        fn.__doc__ = fn.__doc__ + addFromstr + from_fn.__doc__[index+5:]
        return fn

    return docstring_decorator

def sort_and_find(data, cols_unique_vals, x_axis_col, y_axis_col, select_best, criteria):
    
    """
    Finds a sorted list of "y_axis_col" in "data" based on the "criteria". First, this 
    function generates all the combinations of unique cols values and then it creates a 
    list of values for all the combinations. At the end, it sorts this list based on the criteria.  
    
    Args:
    
    data (:obj:`pd.DataFrame` ): The data to be sorted.
    
    cols_unique_vals (:obj:`list`): A list of columns of important hyperparameters 
    with their unique values.
    
    x_axis_col (:obj:`str`): Column name with the values to be plotted on the x axis.

    y_axis_col (:obj:`str`): Column name with the values to be plotted on the y axis.
    
    criteria (:obj:`str`): criteria to sort the list. There are three choices, 
    (i) max, (ii) min, and (iii)mean.
    
    Return:
    
    sorted_list (:obj:`str`): 
    """
 
    value_for_combi=[]
    for comb in generate_kwargs(cols_unique_vals):
        data_copy = data.copy()

        for k, v in comb.items():
            data_copy = data_copy[data_copy[k]==v]

        value_for_combi.append([comb,
                           data_copy[x_axis_col],
                           data_copy[y_axis_col], 
                           data_copy[y_axis_col].mean(),
                           data_copy[y_axis_col].max(), 
                           data_copy[y_axis_col].min()])
        del data_copy
        gc.collect()
        
    if select_best:
        
        if(criteria == 'mean'):
            sorted_list = sorted(value_for_combi, key=lambda x: x[3], reverse=True)

        elif(criteria == 'max'):
            sorted_list = sorted(value_for_combi, key=lambda x: x[4], reverse=True)

        elif(criteria == 'min'): 
            sorted_list = sorted(value_for_combi, key=lambda x: x[5], reverse=True)
    
    return sorted_list

def set_default_vals(num_graphs):

    """
    Sets the default values for font size, line widths, markers etc.
    
    Args:
    
    num_graphs (:obj:`int` ): num of num_graphs to be plotted with maximum value of 4.
    
    """
    
    if num_graphs==1:
        fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(18,9))
        axes_unrolled = [axes]
        mpl.rcParams['lines.linewidth'] = 3.0
        mpl.rcParams['lines.markersize'] = 12.0
        mpl.rcParams['lines.markeredgewidth'] = 1.0
        mpl.rcParams['xtick.labelsize'] = 27
        mpl.rcParams['ytick.labelsize'] = 27
        mpl.rcParams['axes.labelsize'] = 34.5
        mpl.rcParams['font.size']=23
    
    elif num_graphs==2:
        
        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(18, 7))
        axes_unrolled = axes.reshape(1,2)[0]
        mpl.rcParams['lines.linewidth']=6.0
        mpl.rcParams['lines.markersize']=12.0
        mpl.rcParams['lines.markeredgewidth']=1.0
        mpl.rcParams['xtick.labelsize']=27
        mpl.rcParams['ytick.labelsize']=27
        mpl.rcParams['axes.labelsize']=34.5
        mpl.rcParams['font.size']=23
    
    else:
        
        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(33, 22))
        axes_unrolled = axes.reshape(1,4)[0]
        mpl.rcParams['lines.linewidth'] = 6.6
        mpl.rcParams['lines.markersize'] = 22
        mpl.rcParams['lines.markeredgewidth'] = 2.2
        mpl.rcParams['xtick.labelsize'] = 44
        mpl.rcParams['ytick.labelsize'] = 44
        mpl.rcParams['axes.labelsize'] = 44
        mpl.rcParams['font.size'] = 33.2
    
    return fig, axes, axes_unrolled

def get_default_legend_pos(num_graphs, axes_index=None):
    
    """
    Sets the default values where legends could be placed.
    
    Args:
    
    num_graphs (:obj:`int` ): num of num_graphs to be plotted with maximum value of 4.
    
    axes_index (:obj:`int` , `optional`, defaults to None ): In the case of multiple, 
    setting changed depending on index of axes.
    
    """
    
    if num_graphs==1:
        kwargs={'bbox_to_anchor':(-0.4, 1.00), 'loc':2, 'borderaxespad':0.}
    elif num_graphs==2:
        x_displacement = -1.0 if axes_index%2==0 else 1.0
        kwargs={'bbox_to_anchor':(x_displacement, 1.00), 'loc':2, 'borderaxespad':0.}
    else:
        x_displacement = -0.7 if axes_index%2==0 else 1.0
        kwargs={'bbox_to_anchor':(x_displacement, 1.00), 'loc':2, 'borderaxespad':0.}
    
    return kwargs

def plot_in(axes, axes_index = 0, totplots=1, data=None,  data_to_compare=None, x_axis_col='epoch', 
            y_axis_col="eval_mc", select_best=5, criteria='max', cols_to_find=["w_ramprate"], 
            dis_col='l_fr', dis_val=False, data_to_compare_lb='sup_stats'):
    """
    Main plotting function.
    
    Args:
    
    axes (:obj:`matplotlib.pyplot.axes` ): axes object to plot.
    
    axes_index (:obj:`int`, `optional`, defaults to None ): In the case 
    of multiple, setting changed depending on index of axes.
    
    totplots (obj: int): Total number of plots to be plotted.
    
    data (:obj:`pd.DataFrame` ): Data to sort from.
    
    data_to_compare (:obj:`pd.DataFrame` ): Data to compare with 
    the sorted results. For example, purely supervised results.
    
    x_axis_col (:obj:`str`, `optional`, defaults to 'epoch' ): Column name 
    with the values to be plotted on the x axis.
    
    y_axis_col (:obj:`str`, `optional`, defaults to 'eval_mc' ): Column name 
    with the values to be plotted on the y axis.
    
    select_best (:obj:`int`, `optional`, defaults to 5 ): The number of plots 
    to be made based out of the sorted list.
    
    criteria (:obj:`str`, `optional`, defaults to 'max' ): Criteria to sort the 
    list. There are three choices, (i) max, (ii) min, and (iii)mean.
    
    cols_to_find (:obj:`list`, `optional`, defaults to ['w_ramprate'] ): The list of 
    column names which will analysed to find the best of them based the sorting criteria.
    
    dis_col (:obj:`str`, `optional`, defaults to 'l_fr' ): The dicriminatory column name. 
    This would be column name along which the graphs would be divided along the subplots.
    
    dis_val (:obj:`int` or 'float', `optional`, defaults to 'None' ): This is only valid
    if the 'dis_col' is not None. This is used when a certain unique value of discriminatory 
    column is plotted.
    
    data_to_compare_lb (:obj:`str`, `optional`, defaults to 'sup_stats' ): Label name for 
    the data_to_compare plot. 
    
    """
 
    cols_unique_vals = { v : data[v].unique() for v in cols_to_find }
    sorted_list = sort_and_find(data, cols_unique_vals, x_axis_col, y_axis_col, select_best, criteria)

    plots_range = min(select_best, len(sorted_list))
    
    for i in range(plots_range):
        label_name = '\n'.join([str(k)+'='+str(v) for k, v in sorted_list[i][0].items()])
        axes.plot(sorted_list[i][1], sorted_list[i][2], label=label_name)

    if data_to_compare is not None:
        if dis_val:
            data_2c_dis = data_to_compare[data_to_compare[dis_col] == dis_val]
            axes.plot(data_2c_dis[x_axis_col], data_2c_dis[y_axis_col], label = data_to_compare_lb,linestyle='--')        
        else:
            axes.plot(data_to_compare[x_axis_col], data_to_compare[y_axis_col], label = data_to_compare_lb, linestyle='--')
    
    
    if totplots==3 and axes_index>=1:
        axes.set_xlabel(x_axis_col)
    elif totplots==4 and axes_index>=2:
        axes.set_xlabel(x_axis_col)
    elif totplots<2:
        axes.set_xlabel(x_axis_col)
        
    if dis_val: axes.set_title(dis_col+'='+str(dis_val))
    if axes_index%2==0: axes.set_ylabel(y_axis_col)

@add_end_args(plot_in)
def plot_with_discriminator(dis_col, save_png, data=None, *args, **kwargs):

    """
    Plotter if discriminatory column is specified.
    
    Args:
    
    dis_col (:obj:`str` ): The dicriminatory column name. This would be column
    name along which the graphs would be divided along the subplots.
    
    save_png (:obj:`str` ): Whether to save png of results or not. If the 
    value of save_png is not None then it would save the image with name of 
    string value set in save_png.
    
    kwargs: remaining dictionary of keyword arguments from the plot_in function.

    """
    totplots = len(data[dis_col].unique())
    
    fig, axes, axes_unrolled = set_default_vals(totplots)    
    
    for dis_index, dis_val in enumerate(data[dis_col].unique()):
        
        data_dis = data[data[dis_col]==dis_val]
                
        plot_in(axes_unrolled[dis_index], axes_index=dis_index, totplots=totplots, 
                data=data_dis, dis_val = dis_val, *args, **kwargs)
                
        axes_unrolled[dis_index].legend(**get_default_legend_pos(len(data[dis_col].unique()), dis_index))
                
    if save_png :plt.savefig(save_png, bbox_inches='tight')


@add_end_args(plot_in)
def simple_plot(save_png, *args, **kwargs):

    """
    Plotter if discriminatory column is not specified.
    
    Args:

    save_png (:obj:`str` ): Whether to save png of results or not. If the value of
    save_png is not None then it would save the image with name of string value set 
    in save_png.
    
    kwargs: remaining dictionary of keyword arguments from the plot_in function.

    """
        
    fig, axes, _ = set_default_vals(1)
        
    plot_in(axes, *args, **kwargs)
    
    axes.legend(**get_default_legend_pos(1))
    
    if save_png : plt.savefig(save_png, bbox_inches= 'tight')

@add_end_args(plot_in)
def sort_and_plot(dis_col=None, save_png='results.png', *args, **kwargs):

    """
    Function to sort the results and plot them depending if discriminatory is specified or not.
    
    Args:
    
    dis_col (:obj:`str`, `optional`, defaults to None ): The dicriminatory 
    column name. This would be column name along which the graphs would be 
    divided along the subplots. If it is None, it will simply plot a sorted values.
    
    save_png (:obj:`str` ): Whether to save png of results or not. If the value of 
    save_png is not None then it would save the image with name of string value set 
    in save_png.
    
    kwargs: remaining dictionary of keyword arguments for the plot_in function.
    
    """
    
    if dis_col is None:
        simple_plot(save_png, *args, **kwargs)
    else:
        plot_with_discriminator(dis_col, save_png, *args, **kwargs)



from setuptools import setup, find_packages

setup(
    name="ssfinetuning",
    version="0.2.1", 
    author="Ankush Checkervarty",
    author_email="ankushc48@gmail.com",
    description="A package for fine tuning of pretrained NLP transformers using Semi Supervised Learning",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    keywords="NLP transformers huggingface deep learning pytorch",
    url="https://github.com/Ankush7890/ssfinetuning",
    package_dir={"": "src"},
    packages=find_packages("src"),
    python_requires=">=3.6.0",
    install_requires=['torch>=1.7','transformers==4.2.2', 'datasets==1.5', 'matplotlib>=3.1.3'],
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    test_suite='tests'
)



import torch
from datasets import concatenate_datasets, DatasetDict
from datasets import Dataset as Dataset_hf
import numpy as np
from torch.utils.data import Dataset
import gc
from collections import defaultdict
import pandas as pd
import inspect

def extract_keys(function, kwargs, remove_from_orignal=True):

    """ 
    Function which extract the keys of "function" from "kwargs" by inspecting the signature of "function".
    
    """
    
    signature = inspect.signature(function)
    function_keys = signature.parameters.keys()
    
    if remove_from_orignal:
        kwargs_fn = {k:kwargs.pop(k) for k in function_keys if k in kwargs}
    else:
        kwargs_fn = {k:kwargs[k] for k in function_keys if k in kwargs}
        
    return kwargs_fn



def match_with_batchsize(lim, batchsize):
    """ 
    Function used by modify_datasets below to match return the integer closest to lim 
    which is multiple of batchsize, i.e., lim%batchsize=0.
    
    """

    if lim%batchsize==0:
        return lim        
    else:
        return lim - lim%batchsize
    


def modify_datasets(dataset, labeled_fr=0.5, model_type='TriTrain',labeled1_frac=0.33, train_key="train",
                    label_column='label', unlabeled_labels=-1, batchsize=16):
    
    """ 
    Function to modify pyarrow based datasets (huggingface dataset) for testing 
    at different fraction of labeled vs unlabeled data.
    
    Args:
        
    dataset (:obj:`dataset.DatasetDict`): Dictionary containing training and validation datasets.
    
    labeled_frac (:obj:`float`): Fraction of training dataset to be kept as labeled dataset and 
    rest will be divided as unlabeled dataset.

    model_type (:obj:`str`): Semi supervised model type.

    labeled1_frac (:obj:`float`): In the case of CoTraining and TriTraining model_type, this 
    is the fraction given to the first two models (m1 and m2) after being divided by labeled_fr. 
    Rest is given to model 3. For example, labeled1_frac=0.33, m1 and m2 gets 0.33 
    and m3 gets (1-2*0.33) 
    
    train_key (:obj:`str`):  Key value of where training data is accessed.

    label_column (:obj:`str`):  Key value of where columns for labels.

    unlabeled_labels (:obj:`int`):  Value to be assigned to the unlabeled dataset labels, 
    required for Pi, TemporalEnsemble, and MeanTeacher as they need to know which ones 
    are unlabeled examples.
    
    batchsize (:obj:`int`):  Batch size used during training.

    Return:
    dataset.DatasetDict object with labeled and unlabeled data. 
    """

    
    dataset_train = dataset[train_key].shuffle()
    num_rows = dataset_train.num_rows
    
    lab_lim = match_with_batchsize(int(labeled_fr*num_rows), batchsize)
    unlab_start = num_rows - match_with_batchsize(num_rows-lab_lim, batchsize)
    dataset['labeled'] = dataset_train.select(np.arange(0, lab_lim))        
    dataset['unlabeled'] = dataset_train.select(np.arange(unlab_start, num_rows))
    
    if model_type=='PiModel' or model_type=='TemporalEnsemble' or model_type == 'MeanTeacher':
        
        dataset['unlabeled'] = dataset['unlabeled'].map(lambda x:{label_column:unlabeled_labels})
        
        del dataset[train_key]
        
        dataset[train_key] = concatenate_datasets([dataset['labeled'], dataset['unlabeled']])
        
        del dataset['unlabeled']
        gc.collect()
        return DatasetDict(dataset)
    
    elif model_type=='CoTrain':
        
        dataset['labeled1'] = dataset_train.select(np.arange(0, int(labeled1_frac*lab_lim)))
        dataset['labeled2'] = dataset_train.select(np.arange(int(labeled1_frac*lab_lim), lab_lim))
        dataset['unlabeled'].remove_columns_(label_column)

    elif model_type=='TriTrain':
        
        dataset['labeled1'] = dataset_train.select(np.arange(0, int(labeled1_frac*lab_lim)))
        dataset['labeled2'] = dataset_train.select(np.arange(int(labeled1_frac*lab_lim), int(2*labeled1_frac*lab_lim)))
        dataset['labeled3'] = dataset_train.select(np.arange(int(2*labeled1_frac*lab_lim), lab_lim))
        dataset['unlabeled'].remove_columns_(label_column)
    
    else:
        
        dataset['labeled'] = dataset_train.select(np.arange(0, lab_lim))        
        dataset['unlabeled'].remove_columns_(label_column)
    
    del dataset[train_key]
    gc.collect()
    return DatasetDict(dataset)

def dic_to_pandas(history, loss_key='eval_loss', accuracy_measure='eval_matthews_correlation'):
    """ 
    Function to convert the list of dictionary to pandas DataFrame which is easier for the plotting
    function in plotting utils to handle
    
    Args:
    
    history (:obj:`list`): A list of history dictionaries. It is basically 
    transformer.TrainerState() at different hyperparameters analysed. 
    
    loss_key (:obj:`str`): The key to look for. In the case of analysis of evaluation history, 
    the key is 'eval_loss'. In the case of analysis of training history the key is 'train_loss'
   
    accuracy_measure (:obj:`str`): Name of the metric used during evaluation.
    
    """

    enteries_in_old_dics=0
    dic_state=0
    
    stats = defaultdict(list)
    
    for state in history:
        
        for dics in state:
            
            if type(dics)==dict and loss_key in dics.keys():
                
                stats['epoch'].append(dics["epoch"])
                stats['step'].append(dics["step"])
                stats[loss_key].append(dics[loss_key])
                
                if accuracy_measure: stats[accuracy_measure].append(dics[accuracy_measure])
                
                enteries_in_old_dics += 1
            
            elif type(dics)==list or type(dics)==tuple:
                for _ in range(enteries_in_old_dics - dic_state):
                    stats[dics[0]].append(dics[1])
        
        dic_state = enteries_in_old_dics
    
    return pd.DataFrame(data=stats)

class SimpleDataset(Dataset):
    """
    A simple dataset for utilities in Co-Training and Tri-Training.   

    Args:
        
    dataset (:obj:`Union[SimpleDataset, dataset]`): Can be SimpleDataset object 
    or pyarrow based dataset object.

    Class attributes:
    
     -**original_len**: Length of the dataset at the instantiation.
     
     -**to_append_dic**: The dictionary used in appending unlabeled examples in dataset.
     
     -**batch_masks**: This dictionary keeps track of the unlabeled examples which are 
     removed and inserted in the dataset during appending procedure.
     
    """
   
    def __init__(self, dataset):
        
        if isinstance(dataset, SimpleDataset):
            self.dataset = dataset.dataset
        else:
            self.dataset = dataset
        
        self.original_len = len(self.dataset)
        
        self.to_append_dic = defaultdict(list)
        self.batch_masks = defaultdict(torch.tensor)
                    
    def append(self, ul_data, mask = None, batch_index = None):

        """ 
        Function used during the appending procedure. 
        
        Args:
        
        ul_data (:obj:`torch.FloatTensor`): Unlabeled data batch.
        
        mask (:obj:`torch.BoolTensor`): Mask of the data object which are going to accepted 
        from the batch. This object also helps in keeping track of the examples which are inserted. 
        
        batch_index (:obj: ´int´): Index of the batch of unlabeled data. To be used by batch_masks
        dictionary. 
        
        Return:
        mask_change.sum() (:obj: ´int´): Sum of any insertion and deletion of examples in the dataset. 
        
        """

        mask = torch.ones(ul_data['input_ids'].size()[0]).bool() if mask is None else mask
        
        for k in ul_data:
            to_add = ul_data[k][mask].tolist()
            if len(to_add):
                self.to_append_dic[k] += to_add

        if batch_index is not None:

            mask_change = mask.clone()

            if batch_index not in self.batch_masks.keys():
                self.batch_masks[batch_index] = mask

            else:
                exists = torch.logical_and(self.batch_masks[batch_index], mask)
                mask_change[exists]=False

                changed_to_false = torch.logical_and(self.batch_masks[batch_index], mask==False)
                self.batch_masks[batch_index][changed_to_false]=False
                mask_change[changed_to_false]=True

                self.batch_masks[batch_index] = torch.logical_or(self.batch_masks[batch_index], mask)
            
            return mask_change.sum()            
        
            
    def reset(self):
        """ 
        Resets the dataset to the original length at the instantiation.                 
        """
        self.dataset = self.dataset.select(np.arange(0, self.original_len))
        
    def reformat(self):

        """ 
        After appending using usual list appending, dataset is reformat to 
        huggingface dataset format.
        """

        for k in self.to_append_dic:
            self.to_append_dic[k] = self.dataset[k][:] + self.to_append_dic[k]
        
        if len(self.to_append_dic):
            self.dataset = Dataset_hf.from_dict(self.to_append_dic)
        
        self.to_append_dic.clear()
        gc.collect()
        
        return self 
        
    def __len__(self):
    
        return len(self.dataset)
    
    def __getitem__(self, idx):
        
        return self.dataset[idx]
            
    def extend_length(self, length):
        """ 
        Extends the length of the dataset by randomly repeating length amount of rows.
        """
        
        len_=len(self)
        
        if length<len_:
            raise( 'Should not decrease the length of dataset' )
        
        rand_indices = np.random.randint(len_, size=length - len_)
        columns = self.dataset.format['columns']
        
        additional_data = self.dataset.select(rand_indices)
        self.dataset = concatenate_datasets([self.dataset, additional_data])
        self.dataset.set_format(type=self.dataset.format["type"], columns=columns)
